import dataclasses


@dataclasses.dataclass
class OSLODatatypeComplex:
    name: str = ''
    objectUri: str = ''
    definition: str = ''
    label: str = ''
    usagenote: str = ''
    deprecated_version: str = ''




